﻿using AngleSharp;
using AngleSharp.Css;
using AngleSharp.Dom;
using GeneThomas;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net.Http;
using System.Text;
using System.Text.RegularExpressions;
using System.Threading.Tasks;

class AngleSharpCssSpeedFault
{
    static void Main(string[] args)
    {
        string covidHtml = Download("https://www.worldometers.info/coronavirus/", "coronavirus.html", false);
        using (new TimeOperation("InnerText"))
        {
            Dictionary<string, Dictionary<string, string>> covid = ExtractTable(covidHtml, "#main_table_countries_today", 0);
            Console.WriteLine();
        }
        Console.ReadKey();

        // My InnerText: ended Sat 9 May 20 11:41:02, elapsed 04.6691
        // AngelShaarp.Css IElemnt.InnerText(): ended Sat 9 May 20 11:50:19, elapsed 08:31.9458
    }


    public static void CheckResponseOkAsync(HttpResponseMessage response)
    {
        if (response.IsSuccessStatusCode)
            return;

        string responseString = response.Content.ReadAsStringAsync().Result;
        throw new Exception($"HTTP {(int)response.StatusCode} {response.StatusCode} {response.ReasonPhrase}:\n{responseString}");
    }

    public static Dictionary<string, Dictionary<string, string>> ExtractTable(string html, string tableSelector, int keyColumn)
    {
        //IConfiguration config = Configuration.Default.WithCss();
        //Create a new context for evaluating webpages with the given config

        IConfiguration config = Configuration.Default
            .WithCss()
            .WithRenderDevice(new DefaultRenderDevice
            {
                DeviceHeight = 768,
                DeviceWidth = 1024,
            });

        IBrowsingContext context = BrowsingContext.New(config);
        //Create a virtual request to specify the document to load (here from our fixed string)
        IDocument document = context.OpenAsync(req => req.Content(html)).Result;

        // IHtmlCollection<IElement> 
        IElement table = document.QuerySelectorAll(tableSelector).Single();
        List<string> headers = new List<string>();

        Dictionary<string, Dictionary<string, string>> result = new Dictionary<string, Dictionary<string, string>>(StringComparer.InvariantCultureIgnoreCase);
        foreach (IElement tr in table.QuerySelectorAll("tr"))
        {
            if (headers.Count == 0)
            {
                foreach (IElement th in tr.QuerySelectorAll("th"))
                {
                    //string str = th.Text();
                    // removes <br>

                    string str = InnerText(th);

                    headers.Add(str);
                }
                continue;
            }

            Dictionary<string, string> row = new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase);
            string countryName = "";
            int n = -1;
            foreach (IElement td in tr.QuerySelectorAll("td"))
            {
                ++n;
                if (n == keyColumn)
                {
                    countryName = InnerText(td);
                    continue;
                }
                row[headers[n]] = InnerText(td);
            }

            result[countryName] = row;
            //Console.WriteLine($"{countryName}: {row.Count()} cols");
        }
        return result;
    }
    static string Download(string url, string filename, bool forceDownload)
    {
        if (File.Exists(filename) && !forceDownload)
        {
            return File.ReadAllText(filename);
        }

        Console.WriteLine($"Downloading {url}");
        HttpClient client = new HttpClient();
        HttpResponseMessage response = client.GetAsync(url).Result;
        CheckResponseOkAsync(response);

        StreamContent scontent = (StreamContent)response.Content;
        string text = response.Content.ReadAsStringAsync().Result;
        File.WriteAllText(filename, text);
        return text;
    }

    /// <summary>
    /// Extract text as web browser does,
    /// ＜br＞ -> \n.
    /// Raised https://github.com/AngleSharp/AngleSharp/issues/877
    /// </summary>
    /// <param name="elem"></param>
    /// <returns></returns>
    /// 
    static int NumExtractText = 0;
    static int MaxExtractText = 3003;
    static ConsoleProgressBar bar = new ConsoleProgressBar(MaxExtractText);

    // https://developer.mozilla.org/en-US/docs/Web/API/HTMLElement/innerText
    static string InnerText(IElement elem)
    {
        ++NumExtractText;
        bar.Value = NumExtractText;

        //elem.Text;
        //elem.TextContent

#if false
        // elapsed 08:59.36

        string str = elem.GetInnerText();
        return str;
        
#else
        //string str = elem.InnerText;
        string str = elem.InnerHtml;
        str = Regex.Replace(str, "\\s", " ");
        str = str.Replace("&nbsp;", " ");
        str = str.Replace("<br>", "\n");
        str = Regex.Replace(str, "<[^>]+>", "");  // remove elements
        str = Regex.Replace(str, " +", " ");      // many spaces are one space
        str = Regex.Replace(str, " *\n *", "\n"); // “one <br> two” -> “one\ntwo”
        str = str.Replace("&gt;", ">");
        str = str.Replace("&lt;", "<");
        str = str.Replace("&quot;", "\"");
        str = str.Replace("&apos;", "'");
        str = str.Replace("&amp;", "&");
        // any other html &entity; possible
        str = str.Trim(' '); // leave leading and trailsing \n (<br>)s
        return str;
#endif
    }

}
