﻿using System;
using System.IO;

namespace GeneThomas
{
    // should be static extension methods on Console
    public static class ConsoleExtensions
    {
        public static void ClearLine()
        {
            // ConsoleExtensions.ClearLine()
            Console.Out.ClearLine();
        }

        public static void ClearLine(this TextWriter writer)
        {
            // Console.Error.ClearLine();
            int width;
            if (Console.IsOutputRedirected)
                width = 50;
            else
                width = Console.BufferWidth;
            writer.Write("\r{0}\r", " ".Multiply(width - 1));
        }

        public static bool Prompt(string prompt)
        {
            Console.Write(prompt);
            for (; ; )
            {
                ConsoleKeyInfo key = Console.ReadKey();
                for (; ; )
                {
                    if (key.Modifiers.HasFlag(ConsoleModifiers.Alt))
                        break;
                    if (key.Modifiers.HasFlag(ConsoleModifiers.Control))
                        break;
                    if (key.Key == ConsoleKey.Y)
                    {
                        Console.WriteLine();
                        return true;
                    }
                    if (key.Key == ConsoleKey.N)
                    {
                        Console.WriteLine();
                        return false;
                    }
                    break;
                }
                Console.WriteLine(key.KeyChar);
                Console.Write(prompt);
            }
        }

        public static bool TryGetKey(out ConsoleKeyInfo info)
        {
            if (!Console.KeyAvailable)
            {
                info = new ConsoleKeyInfo();
                return false;
            }
            info = Console.ReadKey();
            return true;
        }

        /// <summary>
        /// When printed should see:
        /// 0.5: ▌
        /// 1.0: █
        /// </summary>
        /// <returns></returns>
        public static string BarTest()
        {
            return "0.5: " + Bar(0.5, 1.0, 1) + "\n1.0: " + Bar(1.0, 1.0, 1);
        }

        /// <summary>
        /// return a string bar graph, e.g: "█████▌";
        /// requries Console.OutputEncoding = Encoding.Unicode;
        /// [defaults to codepage 850];
        /// or full block rendered as half block 
        /// </summary>
        /// <param name="value"></param>
        /// <param name="max"></param>
        /// <param name="width"></param>
        /// <returns></returns>
        public static string Bar(double value, double max, int width)
        {
            //Console.WriteLine("bar:  x█y");
            //Console.WriteLine("hbar: x▌y");
            //Console.WriteLine("█▏"); // full bar and eigth
            // rendered as [?]


            int w = (int)(value / max * width * 8 + 0.5);
            string bar = "█".Multiply(w / 8);
            // █ U+02588 FULL BLOCK
            // ▉ U + 02589 LEFT SEVEN EIGHTHS BLOCK
            // ▊ U + 0258A LEFT THREE QUARTERS BLOCK
            // ▋ U + 0258B LEFT FIVE EIGHTHS BLOCK
            // ▌ U + 0258C LEFT HALF BLOCK
            // ▍ U + 0258D THREE EIGHTHS BLOCK
            // ▎ U + 0258E LEFT ONE QUARTER BLOCK
            // ▏ U + 0258F LEFT ONE EIGHTH BLOCK

            //const string parts   = " ▏▎▍▌▋▊▉";
            //                      01234567
            // printed as [?] in Conmand Prompt
            // with Consalis are not the same height
            // in notepad - must use other fonts

            //const string halves = "    ▌▌▌▌";

            //                    01234567
            //                    000hhhww
            const string parts = "   ▌▌▌██";

            if (w % 8 > 0)
            {
                bar += parts[w % 8];
            }

            return bar;
        }
    }
}
