﻿
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace GeneThomas
{
    // ProgressBar for the Console

    public class ConsoleProgressBar
    {
        public ConsoleProgressBar()
        {
            ResetStart();
        }
        public ConsoleProgressBar(int max) :
        this()
        {
            Maximum = max;
        }

        public double Value
        {
            // the value to show
            get
            {
                return theValue;
            }
            set
            {
                if (value < 0.0 || (HasMaximum && value > max) || theValue == value || value < startValue)
                    return;
                theValue = value;
                Draw();
            }
        }

        public void Inc()
        {
            Value = Value + 1;
        }

        public bool HasMaximum
        {
            get
            {
                return max > 0.0;
            }
        }

        double startValue = 0.0;
        public double StartValue
        {
            get
            {
                return startValue;
            }
            set
            {
                startValue = value;
                theValue = value;
                lastPeriodValue = value;
            }
        }

        public double Maximum
        {
            // the maximum Value will get to
            get
            {
                return max;
            }
            set
            {
                if (value <= 0.0)
                    return;
                max = value;
                if (theValue > max)
                    theValue = max;
                if (HasOutput)
                    Draw();
            }
        }

        /// <summary>
        ///  automatically write output in an update
        /// </summary>
        public bool AutoDraw { get; set; } = true;

        public char BarChar { get; set; } = '█';
        // char to put in bar
        // ∎ ENF OF PROOF not supported on Windoze
        // █ too big in []

        public string StartBar { get; set; } = "▐";
        // "▐"; 
        // "▕"; not on windows
        // "▐";
        // 	Right one eighth block show as ? in CommandPrompt
        // "|";
        // text before the bar

        public string EndBar { get; set; } = "|";
        // "▎";  as ?
        // "▏"; 
        // "▌"; too big
        //" | ";
        // text after the bar

        public void SetAscii()
        {
            StartBar = "|";
            BarChar = '=';
            EndBar = "|";
        }

        public int Width
        {
            // the width of the bar graph
            // does not include leading '['
            // trailing ']' or ToGo and PerSec
            // text
            get
            {
                return maxWidthChars;
            }
            set
            {
                if (value <= 0)
                    return;
                Clear();
                maxWidthChars = value;
                lastOutput = "";
                Draw();
            }
        }

        // @todo: all setted should clear lastOutput
        /// <summary>
        /// see also `StartCarrigeReturn`
        /// </summary>
        public string Prefix { get; set; } = "";

        public bool ShowValue { get; set; } = true;
        // 123 or 123/5678

        public bool ShowMax { get; set; } = true;
        // 123 or 123/5678

        public string ValueSuffix { get; set; } = "";

        public bool ShowBar { get; set; } = true;
        // [==        ]

        public string BarSuffix { get; set; } = "";

        public bool ShowEtta { get; set; } = true;
        // show "3:20 etta" after progress bar

        public bool ShowTotal { get; set; } = false;
        // show "3:20 total" after progress bar

        public bool UseConsole { get; set; } = false;
        // use the console API to get the window width

        /// <summary>
        /// for backwards comatibility
        /// </summary>
        public bool ShowTogo
        {
            get
            {
                return ShowEtta;
            }
            set
            {
                ShowEtta = value;
            }
        }
        public bool ShowAll
        {
            set
            {
                ShowValue = value;
                ShowBar = value;
                ShowEtta = true;
                ShowRate = true;
                ShowTotal = true;
                if (AutoDraw)
                    Draw();
            }
        }

        /// <summary>
        /// show Estimated Time To Arrival
        /// </summary>
        public string EttaSuffix { get; set; } = " ᴇᴛᴛᴀ";
        // text to show after the days/housr/mins/sec to go

        public string TotalSuffix { get; set; } = " total";
        // text to show after the days/housr/mins/sec total

        /// <summary>
        /// for backwards comatibility
        /// </summary>
        public string ToGoSuffix
        {
            get
            {
                return EttaSuffix;
            }
            set
            {
                EttaSuffix = value;
            }
        }

        public bool ShowRate { get; set; } = false;
        // show "230/sec

        public string RateSuffix { get; set; } = "";

        public double Divisor { get; set; } = 1.0;
        // Units to show in /sec text
        // to report units/sec as (units/Divisor)/sec

        public string Unit { get; set; } = "";
        // the unit so show in the /sec
        // e.g:
        // Divisor = 1024
        // Unit = "KiB"

        public string Suffix { get; set; } = "";

        /// <summary>
        /// return the cursor to the left of the screen
        /// before writing output,
        /// defaults to true
        /// </summary>
        public bool StartCarrigeReturn { get; set; } = true;

        public bool EndCarrigeReturn { get; set; } = false;

        /// <summary>
        /// defaults to Console.Out
        /// </summary>
        public TextWriter Writer { get; set; } = null;

        // =========================================================

        private double theValue = 0;
        // Value

        private double max = 0;
        // Maximum, HasMaximum

        private int maxWidthChars = 40;
        // Width
        // width of bar in chars
        // has '[' and ']' around it

        private string lastOutput = "";

        private bool HasOutput
        {
            get
            {
                return lastOutput != "";
            }
        }

        private DateTime start;
        // when we started
        // used to calculate time to-go
        // and rate

        private DateTime lastPeriodStart = DateTime.MinValue;
        // when we last reported the rate

        private double lastPeriodValue = Double.MinValue;
        // value at time lastPeriodStart

        private string lastRateStr = "";

        private TimeSpan minPeriod = TimeSpan.FromSeconds(1);

        public void ResetStart()
        {
            start = DateTime.Now;
            lastPeriodStart = start;
            lastPeriodValue = StartValue;
            theValue = StartValue;
        }

        public override string ToString()
        {
            string barStr = "";
            // [==       ]

            string valueStr = "";
            // 12/345

            string rateStr = "";
            // 34.5 KiB/sec

            string ettaStr = "";
            // 2:30 etta

            string totalStr = "";
            // 5:23 total

            double n = theValue - StartValue;

            if (ShowBar)
            {
                int barValueChars = 0;
                if (HasMaximum)
                    barValueChars = (int)((double)maxWidthChars * n / (max - StartValue));
                barStr = StartBar + BarChar.Multiply(barValueChars) + " ".Multiply(maxWidthChars - barValueChars) + EndBar;
            }

            if (ShowValue)
            {
                valueStr = String.Format($"{theValue / Divisor:N0}");
                if (ShowMax && HasMaximum)
                {
                    if (StartValue == 0)
                        valueStr += String.Format($"/{Maximum / Divisor:N0}");
                    else
                        valueStr += String.Format($"/{StartValue / Divisor:N0}..{Maximum / Divisor:N0}");
                }

            }

            if (n != 0 && HasMaximum)
            {
                DateTime now = DateTime.Now;
                TimeSpan elapsed = now - start;
                double totalTicks = (double)elapsed.Ticks / (n / (max - StartValue));
                TimeSpan total = new TimeSpan((long)totalTicks);

                if (ShowEtta)
                {
                    TimeSpan etta = total - elapsed;
                    if (etta.Days > 0)
                    {
                        ettaStr += etta.Days.PluralToString("day") + ", ";
                        etta -= new TimeSpan(TimeSpan.TicksPerDay * etta.Days);
                    }
                    ettaStr += $@"{etta:hh\:mm\:ss}";
                    //String.Format("{0:d}:{1:d2} to go", togo.Hours, togo.Minutes, to);
                    // TimeSpan does not do 'h', only 'hh'
                }

                if (ShowTotal)
                {
                    if (total.Days > 0)
                    {
                        totalStr += String.Format("{0} days, ", total.Days);
                        total -= new TimeSpan(TimeSpan.TicksPerDay * total.Days);
                    }
                    totalStr += $@"{total:hh\:mm\:ss}";
                    //String.Format("{0:d}:{1:d2} to go", togo.Hours, togo.Minutes, to);
                    // TimeSpan does not do 'h', only 'hh'
                }
            }

            if (ShowRate)
            {
                DateTime now = DateTime.Now;
                if (now > lastPeriodStart + minPeriod || lastPeriodStart == start)
                {
                    // time for an update
                    double diffValue = theValue - lastPeriodValue;
                    double perSec = 0.0;
                    if (diffValue == 0)
                    {
                        if (lastPeriodStart == start)
                            rateStr = "";
                        else
                            rateStr = "stalled";
                    }
                    else
                    {
                        TimeSpan span = now - lastPeriodStart;
                        double secs = ((double)span.Ticks) / TimeSpan.TicksPerSecond;
                        perSec = diffValue / Divisor / secs;
                        rateStr = String.Format("{0:N1}{1}{2}/sec", perSec, Unit == "" ? "" : " ", Unit);
                    }
                    if ((lastPeriodStart == start).Implies(diffValue != 0))
                    {
                        lastPeriodStart = now;
                        lastPeriodValue = theValue;
                    }
                }
                else
                {
                    rateStr = lastRateStr;
                }
                lastRateStr = rateStr;
            }

            StringBuilder sb = new StringBuilder();
            if (BarSuffix != "" && barStr != "")
                barStr += BarSuffix;
            if (ValueSuffix != "" && valueStr != "")
                valueStr += ValueSuffix;
            if (EttaSuffix != "" && ettaStr != "")
                ettaStr += EttaSuffix;
            if (TotalSuffix != "" && totalStr != "")
                totalStr += TotalSuffix;
            if (rateStr != "")
                rateStr += RateSuffix;

            int pos = 0;
            foreach (var s in new String[] { Prefix, valueStr, barStr, rateStr, ettaStr, totalStr })
            {
                if (s == "")
                    continue;
                if (sb.Length != 0)
                {
                    if (pos <= 2)
                        sb.Append(' ');
                    else
                        sb.Append(", ");
                }
                sb.Append(s);
                ++pos;
            }
            if (sb.ToString() != "" && !Suffix.StartsWith(" "))
                sb.Append(" ");
            sb.Append(Suffix);
            string output = sb.ToString();
            return output;
        }

        private void Draw()
        {
            if (!AutoDraw)
                return;
            string output = ToString();
            if (output == lastOutput)
                return;
            lastOutput = output;

            output = (StartCarrigeReturn ? "\r" : "") + output + (EndCarrigeReturn ? "\r" : "");
            TextWriter outf = Writer ?? Console.Out;
            if (UseConsole)
                outf.ClearLine();
            outf.Write(output);
            outf.Flush();
        }

        public void Clear()
        {
            if (UseConsole)
                ConsoleExtensions.ClearLine();
        }
    }
}
