﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace GeneThomas
{
    public static class StringExtensions
    {

        public static string Multiply(this string value, int repeatCount)
        {
            // note: would have implemented as extension operators
            // e.g. "=" * 10
            // but these are not allowed in C#
            // operators must be inside the same class
            //  public static string operator*(int len, string str);
            //  public static string operator *(string str, int len);

            if (repeatCount < 0)
            {
                // as per "".Append('x', 5)
                throw new ArgumentOutOfRangeException("repeatCount", "Multiply: repeatCount must be greater than or equal to 0");
            }
            StringBuilder r = new StringBuilder();
            for (int i = 0; i < repeatCount; ++i)
                r.Append(value);
            return r.ToString();
        }
        public static string Plural(int n, string singular, string plural = null)
        {
            if (n == 1)
            {
                return $"{n:N0} {singular}";
            }
            else
            {
                if (plural == null)
                    plural = singular + "s";
                return $"{n:N0} {plural}";
            }
        }
    }
    public static class BoolExtensions
    {
        /// <summary>
        /// if `this` is true `other` must be true,
        /// should be →→ or ->-> operator as should short circuit
        /// </summary>
        public static bool Implies(this bool thisa, bool other)
        {
            return !thisa || other;
        }
    }

    static public class CharExtensions
    {

        public static string Multiply(this char value, int repeatCount)
        {
            // note: would have implemented as extension operators
            // e.g. 'x' * 10
            // but these are not allowed in C#
            // operators must be inside the same class
            //  public static string operator*(int len, string str);
            //  public static string operator *(string str, int len);

            if (repeatCount < 0)
            {
                // as per "".Append('x', 5)
                throw new ArgumentOutOfRangeException("repeatCount", "Multiply: repeatCount must be greater than or equal to 0");
            }
            StringBuilder r = new StringBuilder();
            for (int i = 0; i < repeatCount; ++i)
                r.Append(value);
            return r.ToString();
        }
    }


    public static class IntExtensions
    {

        /// n.Plural("chicken") => "23 chickens";
        public static string PluralToString(this int n, string singular)
        {
            return StringExtensions.Plural(n, singular);
        }
    }

}
