﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace GeneThomas
{
    public class TimeOperation : IDisposable
    {
        public string Name;
        DateTime Start;
        bool ShowTimes;
        public bool Show = true;
        string NameSpace
        {
            get
            {
                string indentStr = " ".Multiply(Indent);
                if (Name == "")
                    return indentStr;
                else
                    return indentStr + Name + ":";
            }
        }

        static public TextWriter DefaultWriter = Console.Out;

        public TextWriter Writer = DefaultWriter;
        int Indent;
        static int NumTimeOperations = 0;
        static bool ShowStartTimes = false;

        TimeSpan Elapsed;
        public int Skip = 0;
        State TheState = null;

        public class State
        {
            internal int Skipped = 0;
            internal List<TimeSpan> Results = new List<TimeSpan>();

            public TimeSpan Average
            {
                get
                {
                    if (Results.Count() == 0)
                        return TimeSpan.MaxValue;
                    else
                        return new TimeSpan(Results.Select(timespan => timespan.Ticks).Sum() / Results.Count());
                }
            }
            public int Count => Results.Count();
        }
        public TimeOperation(string name, State state) :
            this(name, false, state)
        {
        }
        public TimeOperation(string name = "", bool licoDoesNothingButWasteMyTimeInTheWeekend_showTimes = true, State state = null)
        {
            Name = name;
            Indent = NumTimeOperations;
            Start = DateTime.Now;
            ShowTimes = licoDoesNothingButWasteMyTimeInTheWeekend_showTimes;
            if (ShowTimes && ShowStartTimes)
                Writer.WriteLine($"{NameSpace}started {DateTime.Now}");
            ++NumTimeOperations;
            TheState = state;
        }

        static string Format(TimeSpan span)
        {
            if (span.Days > 0)
                return span.Days.PluralToString("day") + " " + span.ToString(@"hh\:mm\:ss\.ffff");
            if (span.Hours > 0)
                return span.ToString(@"hh\:mm\:ss\.ffff");
            if (span.Minutes > 0)
                return span.ToString(@"mm\:ss\.ffff");
            return span.ToString(@"ss\.ffff");
        }
        public void Dispose()
        {
            Elapsed = DateTime.Now - Start;
            --NumTimeOperations;
            if (TheState != null)
            {
                if (Skip > TheState.Skipped)
                {
                    ++TheState.Skipped;
                }
                else
                {
                    TheState.Results.Add(Elapsed);
                }
            }
            if (Show)
                Writer.WriteLine(ToString());
        }

        public override string ToString()
        {
            Elapsed = DateTime.Now - Start;
            string totals = "";
            if (TheState != null)
            {
                if (Skip > 0)
                    totals = $", skipped {Skip}";
                if (TheState.Results.Any())
                {
                    TimeSpan average = new TimeSpan(TheState.Results.Select(timespan => timespan.Ticks).Sum() / TheState.Results.Count);
                    totals += $", num {TheState.Results.Count}, max {Format(TheState.Results.Max())}, average {Format(average)}, min {Format(TheState.Results.Min())}";
                }
            }
            if (ShowTimes)
                return $"{NameSpace} ended {DateTime.Now}, elapsed {Format(Elapsed)}{totals}";
            return $"{NameSpace} elapsed {Format(Elapsed)}{totals}";
        }
    }
}
